import os
import re
import sys
from abc import ABC
from abc import abstractmethod
from typing import List


RAKAM_SYSTEMS_DIR = os.path.dirname(
    os.path.dirname(os.path.abspath(__file__))  # ingestion  # this file
)
sys.path.append(RAKAM_SYSTEMS_DIR)

from rakam_systems.core import VSFile, NodeMetadata, Node


# This takes big chunks and split them in more granularity

class NodeProcessor(ABC):
    """
    Abstract base class for processing nodes in a VSFile.
    Subclasses should implement the `process` method.
    """

    @abstractmethod
    def process(self, vs_file: VSFile) -> None:
        """
        Process the nodes of a VSFile and modify the VSFile in-place.
        """
        pass

    def _build_node_from_chunk(self, content: str, metadata: NodeMetadata) -> Node:
        chunk_metadata = NodeMetadata(
            source_file_uuid=metadata.source_file_uuid,
            position=metadata.position,  # eg. page_number
            custom=metadata.custom.copy() if metadata.custom else {},
        )
        return Node(content=content, metadata=chunk_metadata)


class CharacterSplitter(NodeProcessor):
    """
    Splits the content of each Node based on a specified number of characters.

    Attributes:
        max_characters(int): The maximum number of characters in each chunk.
        overlap(int): The number of characters to overlap between chunks.

    Methods:
        process(vs_file: VSFile) -> None:
            Split the content of each Node in the VSFile into chunks.
    """

    def __init__(self, max_characters: int = 1024, overlap: int = 20):
        self.max_characters = max_characters
        self.overlap = overlap

    def process(self, vs_file: VSFile) -> None:
        chunked_nodes = []
        for node in vs_file.nodes:
            content = node.content
            metadata = node.metadata
            start_idx = 0
            while start_idx < len(content):
                # Calculate the start and end idxs for the chunk
                end_idx = min(start_idx + self.max_characters, len(content))
                chunk = content[start_idx:end_idx]

                chunk_node = self._build_node_from_chunk(
                    content=chunk, metadata=metadata
                )
                chunked_nodes.append(chunk_node)

                # Update the position for the next chunk, considering the overlap
                start_idx = end_idx - self.overlap

        vs_file.nodes = (
            chunked_nodes  # replace the original nodes with the chunked nodes
        )
        vs_file.processed = True  # mark file as processed


class MarkdownSplitter(NodeProcessor):
    """
    Splits the content of each Node in a VSFile based on Markdown headers.

    Methods:
        process(vs_file: VSFile) -> None:
            Split the content of each Node in the VSFile into chunks based on Markdown headers.
    """

    def process(self, vs_file: VSFile) -> None:
        chunked_nodes = []
        for node in vs_file.nodes:
            content = node.content
            metadata = node.metadata
            markdown_nodes = self._split_content(content, metadata, node)
            chunked_nodes.extend(markdown_nodes)

        vs_file.nodes = (
            chunked_nodes  # replace the original nodes with the chunked nodes
        )
        vs_file.processed = True  # mark file as processed

        return markdown_nodes

    def _split_content(
        self, content: str, metadata: NodeMetadata, node: Node
    ) -> List[Node]:
        markdown_nodes = []
        lines = content.split("\n")
        current_chunk_lines = []
        current_metadata = metadata
        code_block = False

        for line in lines:
            if line.lstrip().startswith("```"):
                code_block = (
                    not code_block
                )  # switch from True to False or vice versa (beginning or end of code block)

            header_match = re.match(r"^(#+)\s(.*)", line)
            if header_match and not code_block:
                if current_chunk_lines:
                    chunk_content = "\n".join(current_chunk_lines).strip()
                    chunk_node = self._build_node_from_chunk(
                        content=chunk_content, metadata=current_metadata
                    )
                    markdown_nodes.append(chunk_node)
                    current_chunk_lines = []

                level = len(header_match.group(1))
                header = header_match.group(2)
                current_metadata = self._update_metadata(metadata, header, level)
            current_chunk_lines.append(line)

        if current_chunk_lines:
            chunk_content = "\n".join(current_chunk_lines).strip()
            chunk_node = self._build_node_from_chunk(
                content=chunk_content, metadata=current_metadata
            )
            markdown_nodes.append(chunk_node)

        return markdown_nodes

    def _update_metadata(
        self, metadata: NodeMetadata, header: str, level: int
    ) -> NodeMetadata:
        new_metadata = metadata.custom.copy() if metadata.custom else {}
        new_metadata.update({"header": header, "level": level})
        return NodeMetadata(
            source_file_uuid=metadata.source_file_uuid,
            position=metadata.position,  # eg. page_number
            custom=new_metadata,
        )


if __name__ == "__main__":  # example usage
    vs_file = VSFile("example.md")
    node = Node(
        content="# Header 1\n\nSome content\n\n## Header 2\n\nMore content",
        metadata=NodeMetadata(source_file_uuid=vs_file.uuid, position=0),
    )
    vs_file.nodes = [node]

    splitter = MarkdownSplitter()
    splitter.process(vs_file)



# components/__init__.py

# Importing using absolute imports
from rakam_systems.components.data_processing.content_extractors import ContentExtractor, SimplePDFParser, PDFContentExtractor, URLContentExtractor, JSONContentExtractor
from rakam_systems.components.data_processing.data_processor import DataProcessor
from rakam_systems.components.data_processing import utils
# Define what is available when importing * from this package
__all__ = [
    "ContentExtractor", "SimplePDFParser", "PDFContentExtractor", "URLContentExtractor", "JSONContentExtractor",
    "DataProcessor", 
    "utils"
    ]


from rakam_systems.components.agents.actions import Action


class MockAgent:
    """
    A mock version of the Agent class to simulate the behavior of the actual Agent.
    """
    def __init__(self):
        self.actions = {}
        self.state = {}

    def add_action(self, action_name: str, action: Action):
        self.actions[action_name] = action

    def choose_action(self, input: str):
        """
        Select the appropriate action based on the input.
        For simplicity, we just return the action associated with the input.
        """
        return self.actions.get(input, None)

    def execute_action(self, action_name: str, **kwargs):
        action = self.choose_action(action_name)
        if action:
            return action.execute(**kwargs)
        return None


class MockAction(Action):
    """
    Mock version of Action class to simulate action execution.
    """
    def __init__(self, agent, **kwargs):
        self.agent = agent

    def execute(self, **kwargs):
        """
        Return a mock response simulating action execution.
        """
        return f"Mock action executed with args: {kwargs}"



import sys
import os

# Add the root directory of rakam_systems to the path
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import pytest
import numpy as np
from rakam_systems.core import Node, NodeMetadata, VSFile
from rakam_systems.components.vector_search.vector_store import VectorStore
from sentence_transformers import SentenceTransformer

@pytest.fixture
def vector_store_fixture(tmp_path):
    """
    Fixture to initialize the VectorStore with a temporary path and real embeddings.
    """
    # Initialize VectorStore
    base_index_path = tmp_path / "vector_store"
    vector_store = VectorStore(base_index_path=str(base_index_path), embedding_model="sentence-transformers/all-MiniLM-L6-v2", initialising=True)

    return vector_store

def test_init_vector_store(tmp_path):
    """
    Test the initialization of the VectorStore.
    """
    base_index_path = tmp_path / "vector_store"
    vector_store = VectorStore(base_index_path=str(base_index_path), embedding_model="sentence-transformers/all-MiniLM-L6-v2", initialising=True)

    # Check if the base index path was created
    assert os.path.exists(base_index_path)

def test_predict_embeddings(vector_store_fixture):
    """
    Test the prediction of embeddings for a given query.
    """
    query = "test query"
    embeddings = vector_store_fixture.predict_embeddings(query)

    assert isinstance(embeddings, np.ndarray)
    assert embeddings.shape[1] == 384  # Shape should match the embedding size

def test_create_collection_from_files(vector_store_fixture, tmp_path):
    """
    Test the creation of a collection from VSFiles.
    """
    vs_file = VSFile(file_path="test_file_path")
    node_metadata = NodeMetadata(source_file_uuid="uuid1", position=1)
    node = Node(content="This is a test content", metadata=node_metadata)
    vs_file.nodes = [node]

    collection_name = "test_collection"
    vector_store_fixture.create_collection_from_files(collection_name=collection_name, files=[vs_file])

    # Check if collection exists
    assert collection_name in vector_store_fixture.collections
    store = vector_store_fixture.collections[collection_name]

    # Check if FAISS index and metadata were created
    assert "index" in store
    assert "category_index_mapping" in store
    assert "metadata_index_mapping" in store
    assert "nodes" in store
    assert len(store["nodes"]) == 1

def test_add_nodes_to_collection(vector_store_fixture):
    """
    Test adding nodes to an existing collection.
    """
    vs_file = VSFile(file_path="test_file_path")
    node_metadata = NodeMetadata(source_file_uuid="uuid1", position=1)
    node = Node(content="This is a test content", metadata=node_metadata)
    vs_file.nodes = [node]

    collection_name = "test_collection"
    vector_store_fixture.create_collection_from_files(collection_name=collection_name, files=[vs_file])

    # Add more nodes to the collection
    new_node_metadata = NodeMetadata(source_file_uuid="uuid2", position=2)
    new_node = Node(content="New test content", metadata=new_node_metadata)
    vector_store_fixture.add_nodes(collection_name=collection_name, nodes=[new_node])

    # Check if nodes were added to the collection
    store = vector_store_fixture.collections[collection_name]
    assert len(store["nodes"]) == 2

def test_delete_nodes_from_collection(vector_store_fixture):
    """
    Test deleting nodes from an existing collection.
    """
    vs_file = VSFile(file_path="test_file_path")
    node_metadata = NodeMetadata(source_file_uuid="uuid1", position=1)
    node = Node(content="This is a test content", metadata=node_metadata)
    vs_file.nodes = [node]

    collection_name = "test_collection"
    vector_store_fixture.create_collection_from_files(collection_name=collection_name, files=[vs_file])

    # Delete nodes from the collection
    vector_store_fixture.delete_nodes(collection_name=collection_name, node_ids=[0])

    # Check if nodes were deleted
    store = vector_store_fixture.collections[collection_name]
    assert len(store["nodes"]) == 0

def test_search_in_collection(vector_store_fixture):
    """
    Test searching in an existing collection.
    """
    vs_file = VSFile(file_path="test_file_path")
    node_metadata = NodeMetadata(source_file_uuid="uuid1", position=1)
    node = Node(content="This is a test content", metadata=node_metadata)
    vs_file.nodes = [node]

    collection_name = "test_collection"
    vector_store_fixture.create_collection_from_files(collection_name=collection_name, files=[vs_file])

    # Perform search
    results, suggested_nodes = vector_store_fixture.search(collection_name=collection_name, query="test")

    # Check if search results are returned
    assert isinstance(results, dict)
    assert len(suggested_nodes) > 0

def test_save_collection(vector_store_fixture, tmp_path):
    """
    Test saving the collection after changes.
    """
    vs_file = VSFile(file_path="test_file_path")
    node_metadata = NodeMetadata(source_file_uuid="uuid1", position=1)
    node = Node(content="This is a test content", metadata=node_metadata)
    vs_file.nodes = [node]

    collection_name = "test_collection"
    vector_store_fixture.create_collection_from_files(collection_name=collection_name, files=[vs_file])

    # Simulate saving collection
    vector_store_fixture._save_collection(collection_name=collection_name)

    # Check if collection files were saved to disk
    store_path = os.path.join(vector_store_fixture.base_index_path, collection_name)
    assert os.path.exists(os.path.join(store_path, "category_index_mapping.pkl"))
    assert os.path.exists(os.path.join(store_path, "metadata_index_mapping.pkl"))
    assert os.path.exists(os.path.join(store_path, "nodes.pkl"))
    assert os.path.exists(os.path.join(store_path, "index"))



import mimetypes
import os
import sys
from typing import Dict
from typing import List

RAKAM_SYSTEMS_DIR = os.path.dirname(
    os.path.dirname(os.path.abspath(__file__))  # ingestion  # this file
)
sys.path.append(RAKAM_SYSTEMS_DIR)

from rakam_systems.core import VSFile
from rakam_systems.components.data_processing import (
    PDFContentExtractor,
    JSONContentExtractor,
)
from rakam_systems.components.data_processing.node_processors import MarkdownSplitter


class DataProcessor:
    def __init__(self) -> None:
        self.default_content_extractors: Dict[str, callable] = {
            "application/pdf": PDFContentExtractor(
                parser_name="AdvancedPDFParser", output_format="markdown", persist=True
            ),
            "application/json": JSONContentExtractor(),
        }
        self.default_node_processors = MarkdownSplitter()

    def process_files_from_directory(self, directory_path: str) -> List[VSFile]:
        # Extract content from files
        print(f"Extracting content from files in {directory_path}")
        vs_files = []
        for root, _, filenames in os.walk(directory_path):
            for filename in filenames:
                file_path = os.path.join(root, filename)
                mime_type, _ = mimetypes.guess_type(file_path)

                if mime_type in self.default_content_extractors:
                    content_extractor = self.default_content_extractors[mime_type]
                    vs_files.extend(content_extractor.extract_content(file_path))
                else:
                    print(
                        f"Skipping file {filename} with unsupported MIME type {mime_type}"
                    )

        # Process nodes
        print(f"Processing nodes in {len(vs_files)} files")
        for vs_file in vs_files:
            self.default_node_processors.process(vs_file)

        return vs_files


if __name__ == "__main__":  # example usage
    processor = DataProcessor()
    vs_files = processor.process_files_from_directory("path/to/directory")



import logging
from abc import ABC
from abc import abstractmethod
from typing import Any

import dotenv

from rakam_systems.components.agents.actions import Action
from rakam_systems.custom_loggers import prompt_logger
from rakam_systems.components.base import LLM

dotenv.load_dotenv()

class Agent(ABC):
    def __init__(self, model: str):
        self.llm = LLM(model=model)  # Initialize the LLM with the specified model
        self.state = {}  # Initialize an empty state dictionary
        self.actions = {}  # Dictionary to store actions by name

    def process_state(self, **kwargs) -> dict:
        """
        Build a temporary state based on the input query.

        :param input: The input query string.
        :return: A dictionary representing the state.
        """
        # Example: Building a state based on keyword analysis
        state = {}
        return state

    def add_action(self, action_name: str, action: Action):
        """
        Adds an action to the agent's set of available actions.

        :param action_name: A string name for the action (e.g., 'rag_generation').
        :param action: An instance of an Action subclass.
        """
        self.actions[action_name] = action

    @abstractmethod
    def choose_action(self, input: str, state: Any) -> Action:
        """
        Abstract method to select an action based on input.

        :param input: A string input based on which the action is selected.
        :return: The selected Action instance.
        """
        pass

    def execute_action(self, action_name: str, **kwargs):
        """
        Executes the selected action with the provided arguments.

        :param action_name: The name of the action to execute.
        :param kwargs: Arguments to pass to the action's execute method.
        :return: The result of the action's execute method.
        """
        action = self.choose_action(action_name)
        return action.execute(**kwargs)




from rakam_systems.core import Node, NodeMetadata
from rakam_systems.components.vector_search.vector_store import VectorStore

BASE_INDEX_PATH = "vector_store_index_1"
EMBEDDING_MODEL = "sentence-transformers/all-MiniLM-L6-v2"

def setup_vector_store():
    vector_store = VectorStore(base_index_path=BASE_INDEX_PATH, embedding_model=EMBEDDING_MODEL)
    return vector_store

def create_nodes(texts):
    return [
        Node(content=text, metadata=NodeMetadata(source_file_uuid=f"file_{i}", position=i))
        for i, text in enumerate(texts)
    ]

def search_vector_store(vector_store, query, number=3):
    results, _ = vector_store.search(collection_name="base", query=query, number=number)
    return results

def print_search_results(query, results):
    print(f"Search Results for query '{query}':")
    for id_, (metadata, suggestion_text, distance) in results.items():
        print(f"ID: {id_}, Suggestion: {suggestion_text}, Metadata: {metadata}, Distance: {distance}")
    print("\n")

def main():
    texts = [
        "This is the first document.",
        "Here is another document.",
        "Final document in the list."
    ]

    vector_store = setup_vector_store()
    nodes = create_nodes(texts)
    vector_store.create_from_nodes(nodes=nodes)

    query = "first document"
    results = search_vector_store(vector_store, query)

    print_search_results(query, results)

if __name__ == "__main__":
    main()



import os
import pytest
import shutil
from rakam_systems.components.vector_search.vector_store import VectorStore, VSFile, Node, NodeMetadata


@pytest.fixture
def setup_vector_store():
    # Create a temporary directory for the test
    base_index_path = "test_data"
    embedding_model = "paraphrase-MiniLM-L6-v2"
    
    # Ensure the test directory is clean before starting
    if os.path.exists(base_index_path):
        shutil.rmtree(base_index_path)
    
    os.makedirs(base_index_path)
    
    # Initialize the VectorStore
    vs = VectorStore(base_index_path=base_index_path, embedding_model=embedding_model)
    yield vs

    # Clean up after the test
    shutil.rmtree(base_index_path)


def test_create_collection_from_files(setup_vector_store):
    vs = setup_vector_store

    # Define example VSFiles with nodes and metadata
    vsfiles_example = [
        VSFile("data/1.txt"),
        VSFile("data/2.txt"),
    ]
    
    # Nodes for the first file
    nodes_1 = [
        Node("This is a test", NodeMetadata("1", 1)),
        Node("This is another test", NodeMetadata("1", 2)),
    ]
    # Nodes for the second file
    nodes_2 = [
        Node("This is a test", NodeMetadata("2", 1)),
        Node("This is another test", NodeMetadata("2", 2)),
    ]
    
    # Assign nodes to the VSFiles
    vsfiles_example[0].nodes = nodes_1
    vsfiles_example[1].nodes = nodes_2

    # Create the collection from the VSFiles
    collection_name = "test_collection"
    vs.create_collection_from_files(collection_name, vsfiles_example)

    # Validate that the nodes have been assigned the correct node_id
    nodes = vs.collections[collection_name]["nodes"]
    assert nodes[0].metadata.node_id == 0, "Node ID for the first node should be 0"
    assert nodes[1].metadata.node_id == 1, "Node ID for the second node should be 1"
    assert nodes[2].metadata.node_id == 0, "Node ID for the third node should be 0"
    assert nodes[3].metadata.node_id == 1, "Node ID for the fourth node should be 1"

    print(f"Assigned node IDs: {[node.metadata.node_id for node in nodes]}")


if __name__ == "__main__":
    pytest.main()



import os
import pytest
from rakam_systems.components.base import LLM
import logging
logging.basicConfig(level=logging.DEBUG)


@pytest.fixture
def llm():
    # Assuming OPENAI_API_KEY is set in your environment
    return LLM(model="gpt-4", api_key=os.getenv("OPENAI_API_KEY"))

def test_call_llm(llm):
    sys_prompt = "You are an assistant."
    user_prompt = "Tell me a fun fact about pandas."

    # Make a real API call
    response = llm.call_llm(sys_prompt, user_prompt)

    # Check that the response is a non-empty string
    assert isinstance(response, str)
    assert len(response) > 0
    logging.debug(f"Response: {response}")
 
def test_call_llm_stream(llm):
    sys_prompt = "You are an assistant."
    user_prompt = "Give me three reasons why pandas are interesting."

    # Make a real API call with streaming
    response_stream = llm.call_llm_stream(sys_prompt, user_prompt)
    logging.debug(f"Stream response: {response_stream}")

    response_chunks = [chunk for chunk in response_stream if chunk]  # Filter out None values
    full_response = "".join(response_chunks)

    # Check that the response is a non-empty string
    assert isinstance(full_response, str)
    assert len(full_response) > 0
    logging.debug(f"Streamed response: {full_response}")




import pytest
import mimetypes
from rakam_systems.core import VSFile, NodeMetadata, Node
import uuid


def test_vsfile_initialization():
    # Test initialization of VSFile
    file_path = "/path/to/sample.pdf"
    vs_file = VSFile(file_path)

    assert isinstance(vs_file.uuid, uuid.UUID), "UUID should be auto-generated"
    assert vs_file.file_path == file_path
    assert vs_file.file_name == "sample.pdf"
    assert vs_file.mime_type == mimetypes.guess_type(file_path)[0], "Mime type should be guessed"
    assert vs_file.nodes == [], "Nodes should be an empty list initially"
    assert not vs_file.processed, "Processed should be False initially"


def test_node_metadata_initialization():
    # Test initialization of NodeMetadata
    metadata = NodeMetadata(source_file_uuid="1234", position=1, custom={"author": "test_user"})

    assert metadata.source_file_uuid == "1234"
    assert metadata.position == 1
    assert metadata.custom == {"author": "test_user"}
    assert metadata.node_id is None, "Node ID should be None by default"


def test_node_metadata_str():
    # Test __str__ method of NodeMetadata
    metadata = NodeMetadata(source_file_uuid="1234", position=1, custom={"author": "test_user"})
    expected_str = (
        "NodeMetadata(node_id=None, source_file_uuid='1234', position=1, custom={ author: test_user })"
    )
    assert str(metadata) == expected_str


def test_node_initialization():
    # Test initialization of Node
    metadata = NodeMetadata(source_file_uuid="1234", position=1, custom={"author": "test_user"})
    node = Node(content="This is a test node content", metadata=metadata)

    assert node.content == "This is a test node content"
    assert node.metadata == metadata
    assert node.embedding is None, "Embedding should be None initially"





import pytest
from rakam_systems.components.agents.mock_agent import MockAgent, MockAction
from rakam_systems.components.agents.actions import (
    TextSearchMetadata, 
    ClassifyQuery, 
    RAGGeneration, 
    GenericLLMResponse
)
import pandas as pd


@pytest.fixture
def mock_agent_fixture():
    """
    Fixture to create a mock agent.
    """
    return MockAgent()

@pytest.fixture
def text_search_metadata_action(mock_agent_fixture):
    """
    Fixture for setting up TextSearchMetadata action.
    """
    text_items = pd.Series(["item1", "item2", "item3"])
    metadatas = pd.Series([{"meta1": "data1"}, {"meta2": "data2"}, {"meta3": "data3"}])
    return TextSearchMetadata(mock_agent_fixture, text_items, metadatas)

@pytest.fixture
def classify_query_action(mock_agent_fixture):
    """
    Fixture for setting up ClassifyQuery action.
    """
    trigger_queries = pd.Series(["query1", "query2", "query3"])
    class_names = pd.Series(["class1", "class2", "class3"])
    return ClassifyQuery(mock_agent_fixture, trigger_queries, class_names)

@pytest.fixture
def rag_generation_action(mock_agent_fixture):
    """
    Fixture for setting up RAGGeneration action.
    """
    return RAGGeneration(mock_agent_fixture, sys_prompt="System prompt", prompt="Prompt", vector_stores=[])

@pytest.fixture
def generic_llm_response_action(mock_agent_fixture):
    """
    Fixture for setting up GenericLLMResponse action.
    """
    return GenericLLMResponse(mock_agent_fixture, sys_prompt="System prompt", prompt="Prompt")




def test_classify_query_execution(classify_query_action):
    """
    Test execution of ClassifyQuery action.
    """
    class_name, trigger_query = classify_query_action.execute(query="query1")
    
    assert class_name == "class1"
    assert trigger_query == "query1"




import mimetypes
import uuid
from typing import Any
from typing import Dict
from typing import List
from typing import Optional


class VSFile:
    """
    A data source to be processed. Its nodes will become entries in the VectorStore.
    """

    def __init__(self, file_path: str) -> None:
        self.uuid: str = uuid.uuid4()
        self.file_path: str = file_path
        self.file_name: str = file_path.split("/")[-1]
        self.mime_type, _ = mimetypes.guess_type(self.file_path)
        self.nodes: Optional[Node] = []
        self.processed: bool = (
            False  # whether the nodes of this file have been processed
        )


class NodeMetadata:
    def __init__(
        self, source_file_uuid: str, position: int, custom: dict = None
    ) -> None:
        self.node_id: Optional[int] = None
        self.source_file_uuid: str = source_file_uuid
        self.position: Optional[int] = position  # page_number
        self.custom: Optional[Dict] = custom

    def __str__(self) -> str:
        custom_str = ", ".join(
            f"{key}: {value}" for key, value in (self.custom or {}).items()
        )
        return (
            f"NodeMetadata(node_id={self.node_id}, source_file_uuid='{self.source_file_uuid}', "
            f"position={self.position}, custom={{ {custom_str} }})"
        )


class Node:
    """
    A node with content and associated metadata.
    """

    def __init__(self, content: str, metadata: NodeMetadata) -> None:
        self.content: str = content
        self.metadata: Optional[NodeMetadata] = metadata
        self.embedding: Optional[Any] = None

    def __str__(self) -> str:
        return f"Node(content='{self.content[:30]}...', metadata={self.metadata})"



import logging
from abc import ABC
from abc import abstractmethod
from typing import Any

import dotenv
import pandas as pd
from typing import List
import os
from rakam_systems.core import Node
from rakam_systems.core import NodeMetadata
from rakam_systems.custom_loggers import prompt_logger
from rakam_systems.components.base import LLM
from rakam_systems.components.vector_search import VectorStore

dotenv.load_dotenv()

# Abstract Action class
class Action(ABC):
    @abstractmethod
    def __init__(self, agent, **kwargs):
        pass

    @abstractmethod
    def execute(self, **kwargs):
        pass

class TextSearchMetadata(Action):
    def __init__(
        self,
        agent,
        collections: dict = {},
        embedding_model: str = "all-MiniLM-L6-v2",
        vector_store_path: str = "temp_path"
    ):
        self.agent = agent
        self.collections = collections
        self.vector_store_path = vector_store_path

        self.embedding_model = embedding_model

        # Initialize the VectorStore
        self.vector_store = self.initialize_vector_store()

    def initialize_vector_store(self) -> VectorStore:
        if not self.vector_store_path:
            # Check the format of collections
            if self._validate_collections_format(self.collections):
                # Create directory if it doesn't exist
                os.makedirs(self.vector_store_path, exist_ok=True)
                # Build the vector store
                return self._build_vector_store()
            else:
                raise ValueError("Invalid collections format.")
        else:
            # Load the existing vector store
            vector_store = VectorStore(
                base_index_path=self.vector_store_path,
                embedding_model=self.embedding_model,
                initialising=False
            )
            vector_store.load_vector_store()

            # Verify collections if provided
            if self.collections:
                self._verify_and_build_collections(vector_store)

            return vector_store

    def _validate_collections_format(self, collections: dict) -> bool:
        # Validate that collections is a dictionary with the correct structure
        for _, data in collections.items():
            if not isinstance(data, dict) or "triggers" not in data or "metadata" not in data:
                return False
        return True

    def _verify_and_build_collections(self, vector_store: VectorStore):
        # Verify that all collections are present in the vector store
        for collection_name, data in self.collections.items():
            if collection_name not in vector_store.collections:
                # Build missing collections
                self._build_collection(vector_store, collection_name, data)

    def _build_collection(self, vector_store: VectorStore, collection_name: str, data: dict):
        # Create nodes from trigger queries and metadata
        nodes = []
        for query, metadata in zip(data["triggers"], data["metadata"]):
            node_metadata = NodeMetadata(
                source_file_uuid=query, position=None, custom=metadata
            )
            node = Node(content=query, metadata=node_metadata)
            nodes.append(node)

        # Use the create_collection_from_nodes method to build the index
        vector_store.create_collection_from_nodes(collection_name=collection_name, nodes=nodes)

    def _build_vector_store(self) -> VectorStore:
        # Initialize VectorStore
        vector_store = VectorStore(
            base_index_path=self.vector_store_path,
            embedding_model=self.embedding_model,
            initialising=True
        )

        # Build all collections
        for collection_name, data in self.collections.items():
            self._build_collection(vector_store, collection_name, data)

        return vector_store
    def execute(self, query: str, collection: str) -> list:
        """
        Classifies the query by finding the closest match in the FAISS index.
        """
        # Perform the search using the vector store
        valid_suggestions, _ = self.vector_store.search(
            collection_name=collection, query=query, number=2
        )

        # print("Successfully classified query:", query, "\nFound : ", valid_suggestions)
        return valid_suggestions

class ClassifyQuery(Action):
    def __init__(
        self,
        agent,
        trigger_queries: pd.Series,
        class_names: pd.Series,
        embedding_model: str = "all-MiniLM-L6-v2",
    ):
        self.agent = agent
        self.trigger_queries = trigger_queries
        self.class_names = class_names
        self.embedding_model = embedding_model
        self.vector_store = self.build_vector_store(trigger_queries, class_names)

    def build_vector_store(
        self, trigger_queries: pd.Series, class_names: pd.Series
    ) -> VectorStore:
        """
        Builds a VectorStore object from the trigger queries and class names.
        """
        # Create nodes from trigger queries and class names
        nodes = []
        for query, class_name in zip(trigger_queries, class_names):
            metadata = NodeMetadata(
                source_file_uuid=query, position=None, custom={"class_name": class_name}
            )
            node = Node(content=query, metadata=metadata)
            nodes.append(node)

        # Initialize VectorStore
        vector_store = VectorStore(
            base_index_path="temp_path", embedding_model=self.embedding_model
        )

        # Use the create_collection_from_nodes method to build the index
        vector_store.create_collection_from_nodes(collection_name="query_classification", nodes=nodes)

        return vector_store

    def execute(self, query: str):
        """
        Classifies the query by finding the closest match in the FAISS index.
        """
        # Perform the search using the vector store
        _, node_search_results = self.vector_store.search(
            collection_name="query_classification", query=query, number=2
        )

        # Extract the matched trigger query and class name
        matched_node = node_search_results[0]
        matched_trigger_query = matched_node.content
        matched_class_name = matched_node.metadata.custom["class_name"]

        print("Successfully classified query:", query)
        return matched_class_name, matched_trigger_query

class RAGGeneration(Action):
    def __init__(
        self,
        agent,
        sys_prompt: str,
        prompt: str,
        vector_stores: List[VectorStore],
        vs_descriptions: List[str] = None,
    ):
        self.agent = agent
        self.default_sys_prompt = sys_prompt
        self.prompt = prompt
        self.vector_stores = vector_stores
        self.vs_descriptions = vs_descriptions

        self.store_separator = "\n====\n"
        self.result_separator = "\n----\n"

    def execute(self, query, collection_names: List [str], prompt_kwargs: dict = {}, stream: bool = False, sys_prompt: str = None, **kwargs):
        ### --- Vector Store Search --- ###
        formatted_search_results = self._perform_vector_store_search(query, collection_names)

        ### --- Format Prompt --- ###
        formatted_prompt = self.prompt.format(
            query=query,
            search_results=formatted_search_results,
            **prompt_kwargs,
        )

        # Use the provided sys_prompt or fall back to the default
        sys_prompt = sys_prompt or self.default_sys_prompt

        prompt_logger.info(f"\nSYSPROMPT:\n---\n{self.default_sys_prompt}\n---\n")
        prompt_logger.info(f"\nPROMPT:\n---\n{self.prompt}\n---\n")
        prompt_logger.info(
            f"\nFORMATTED PROMPT (RAGGeneration):\n---\n{formatted_prompt}\n---\n"
        )

        ### --- LLM Generation --- ###
        if stream: return self._generate_stream(sys_prompt, formatted_prompt)
        else: return self._generate_non_stream(sys_prompt, formatted_prompt)

    # STREAMING (returns a generator)
    def _generate_stream(self, sys_prompt, formatted_prompt):
        # Call the LLM to generate the final answer in streaming mode
        response_generator = self.agent.llm.call_llm_stream(
            sys_prompt, formatted_prompt
        )
        for chunk in response_generator:
            yield chunk

    # NON-STREAMING (returns a string)
    def _generate_non_stream(self, sys_prompt, formatted_prompt) -> str:
        # Call the LLM to generate the final answer
        answer = self.agent.llm.call_llm(sys_prompt, formatted_prompt)
        return answer

    def _perform_vector_store_search(self, query: str, collection_names: List[str]) -> str:
        """
        Perform a search across all vector stores and format the results.
        """

        formatted_search_results = []

        # If vector_store_descriptions is None, use collection_names as descriptions
        if self.vs_descriptions is None:
            self.vs_descriptions = collection_names

        print("self.vector_stores, collection_names, self.vs_descriptions : ", self.vector_stores, collection_names, self.vs_descriptions)
        for store, collection_name, collection_description in zip(self.vector_stores, collection_names, self.vs_descriptions) :
            _, node_search_results = store.search(
                collection_name=collection_name, query=query
            )
            if node_search_results:
                # Format the search results for one vector store
                formatted_search_results.append(f"\n**Source:** {collection_description}\n\n")
                formatted_search_results.append(self._format_search_results(node_search_results))
                formatted_search_results.append(self.store_separator)


        return "".join(formatted_search_results).rstrip(
            self.store_separator + self.result_separator
        )

    def _format_search_results(self, search_results: list) -> str:
        """
        Formats the search results from one vector store into a string.
        """
        if not search_results:
            return f"No results found. {self.result_separator}"

        formatted_results = [
            f"{node.content}{self.result_separator}" for node in search_results
        ]

        return "".join(formatted_results)

class GenericLLMResponse(Action):
    def __init__(self, agent, sys_prompt: str, prompt: str):
        self.agent = agent
        self.default_sys_prompt = sys_prompt
        self.prompt = prompt

    def execute(self, query, prompt_kwargs: dict = {}, stream: bool = False, sys_prompt: str = None, **kwargs):
        ### --- Format Prompt --- ###
        formatted_prompt = self.prompt.format(query=query, **prompt_kwargs)

        # Use the provided sys_prompt or fall back to the default
        sys_prompt = sys_prompt or self.default_sys_prompt

        prompt_logger.info(
            f"\nFORMATTED PROMPT (GenericLLMResponse):\n---\n{formatted_prompt}\n---\n"
        )

        ### --- LLM Generation --- ###
        if stream:
            return self._generate_stream(sys_prompt, formatted_prompt)
        else:
            return self._generate_non_stream(sys_prompt, formatted_prompt)

    # STREAMING (returns a generator)
    def _generate_stream(self, sys_prompt, formatted_prompt):
        # Call the LLM to generate the final answer in streaming mode
        response_generator = self.agent.llm.call_llm_stream(
            sys_prompt, formatted_prompt
        )
        for chunk in response_generator:
            yield chunk

    # NON-STREAMING (returns a string)
    def _generate_non_stream(self, sys_prompt, formatted_prompt) -> str:
        # Call the LLM to generate the final answer
        answer = self.agent.llm.call_llm(sys_prompt, formatted_prompt)
        return answer



import pytest
from rakam_systems.components.agents.mock_agent import MockAgent, MockAction

@pytest.fixture
def mock_agent_fixture():
    """
    Fixture for creating a mock agent instance.
    """
    return MockAgent()

def test_add_action(mock_agent_fixture):
    """
    Test adding actions to the mock agent.
    """
    action = MockAction(mock_agent_fixture)
    mock_agent_fixture.add_action("mock_action", action)
    
    assert "mock_action" in mock_agent_fixture.actions
    assert isinstance(mock_agent_fixture.actions["mock_action"], MockAction)

def test_choose_action(mock_agent_fixture):
    """
    Test choosing an action from the mock agent.
    """
    action = MockAction(mock_agent_fixture)
    mock_agent_fixture.add_action("mock_action", action)
    
    chosen_action = mock_agent_fixture.choose_action("mock_action")
    
    assert chosen_action is not None
    assert isinstance(chosen_action, MockAction)

def test_execute_action(mock_agent_fixture):
    """
    Test executing an action using the mock agent.
    """
    action = MockAction(mock_agent_fixture)
    mock_agent_fixture.add_action("mock_action", action)
    
    result = mock_agent_fixture.execute_action("mock_action", param="test_param")
    
    assert result == "Mock action executed with args: {'param': 'test_param'}"

def test_execute_action_with_no_action(mock_agent_fixture):
    """
    Test executing an action that doesn't exist in the agent.
    """
    result = mock_agent_fixture.execute_action("non_existent_action", param="test_param")
    
    assert result is None
